;;======================================================================;;
;;			GenHSI	 					;;
;;======================================================================;;
;;									;;
;; Program:		Charlieplexing test	;;
;; Code:		Jindra Ficik					;;
;; Platform:		Microchip PIC16F628, 4 Mhz			;;
;; Date:		14.04.2013					;;
;; First release:	14.04.2013					;;
;; LastDate:		14.04.2013					;;
;;									;;
;;======================================================================;;

; Processor PIC 16F628 running at 4MHz.
;
; Revisions:
;
; 14/04/2013	Start of writing code.


; ----- Definitions

#define		__VERNUM	0x01
#define		__VERREV	"A"
#define		__VERDAY	0x14
#define		__VERMONTH	0x04
#define		__VERYEAR	0x13


                list    p=16F628a,r=hex

		errorlevel -302
		errorlevel -306

	        INCLUDE "P16F628A.INC"

                __FUSES _BODEN_ON & _CP_OFF & _PWRTE_ON & _WDT_OFF & _LVP_OFF & _MCLRE_OFF  & _INTRC_OSC_NOCLKOUT


; Port usage
#define		RE_1	PORTA,2		; Rotary encoder first
#define		RE_2	PORTA,3		; Rotary encoder second
#define		RE_B	PORTA,4		; Rotary encoder button


#define		RAMINI0		0x020		; 80 bytes
#define		RAMINI1		0x0A0		; 80 bytes
#define		RAMINI2		0x120		; 48 bytes
#define		RAMINT		0x070		; 16 bytes

CHPORT		equ		PORTB		; port for charlieplexing
CHTRIS		equ		TRISB		; tri state for charlieplexing

; ----- Macros

SEL_BANK_0:	macro	
		bcf	STATUS,RP0		; macros for data access
		bcf	STATUS,RP1
		endm

SEL_BANK_1:	macro	
		bsf	STATUS,RP0
		bcf	STATUS,RP1
		endm

SEL_BANK_2:	macro	
		bcf	STATUS,RP0
		bsf	STATUS,RP1
		endm

SEL_BANK_3:	macro	
		bsf	STATUS,RP0
		bsf	STATUS,RP1
		endm

#define		SEL_IBANK_0	bcf	STATUS,IRP

#define		SEL_IBANK_1	bsf	STATUS,IRP

#define		DNOP		goto	$+1


; ----- Constant values

FXTAL		equ	D'4000000'		; Crystal frequency


RA_TRIS     	equ     0xFF			; all input
RB_TRIS     	equ     0xFF			; all input
RA_INI      	equ     0x00			; TXD, CTS high
RB_INI      	equ     0x00			; 

INTC_INI	equ	0xC0	; PEIE - Prhepial int enable, global int enable
PIE1_INI	equ	0x02	; TMR2IE - Timer 2 int enable (every 1ms)
T2CON_INI equ	0x00	; Prescaler 1:1, Timer 2 disable, must be enabled on demand
PR2_INI equ	0xF9	; 249 dec = 250 instruction = 0.25ms without prescaler
;PR2_INI equ	0x7C	; 124 dec = 125 instruction = 1/8ms without prescaler
;PR2_INI equ	0x52	; 82 dec = 83 instruction ~ 1/12ms without prescaler
;PR2_INI equ	0x3E	; 62 dec = 63 instruction ~ 1/16ms without prescaler
;T2CON_INI equ	0x01	; Prescaler 1:4, Timer 2 disable, must be enabled on demand
;PR2_INI equ	0xF9	; 249 dec = 250 instruction = 1ms with prescaler
;PR2_INI equ	0x7C	; 124 dec = 125 instruction = 0.5ms with prescaler
;PR2_INI equ	0x3F	; 63 dec = 64 instruction = 1ms with prescaler
OPTION_INI	equ	0x82	; Option reg: timer prescaler 1:4, pull-ups
			; Port B Pull-up:       	1   disabled
			; External int edge:    	0   falling
			; Timer trigger source: 	0   clock
			; Timer trigger edge:   	0   rising
			; Prescaler assignment: 	0   Timer0
			; Prescaler value:	010   1:8


; ----- Variables

; --- Top on all banks Section

INT_W		equ	RAMINT+0x00		; Interrupt variables. Last RAM locations
INT_STAT	equ	RAMINT+0x01
INT_PCLATH	equ	RAMINT+0x02
INT_FSR		equ	RAMINT+0x03

EEDATA0		equ	RAMINT+0x04		; EEPROM shadow variables
EEADR0		equ	RAMINT+0x05

; --- Bank 0

NXT_PORT	equ	RAMINI0+0x00		; Data for next port character
NXT_TRIS	equ	RAMINI0+0x01		; Data for next tristate character
BIT_CNT	equ	RAMINI0+0x02		; Counter of bits (pixels)
MSCNT	equ	RAMINI0+0x03		; Counter of miliseconds (decreased every 1ms)
FLAGS	equ	RAMINI0+0x04		; various flags

BALLX	equ	RAMINI0+0x05		; arkanoid ball position X
BALLY	equ	RAMINI0+0x06		; arkanoid ball position Y
BALLDX	equ	RAMINI0+0x07		; arkanoid ball difference X
BALLDY	equ	RAMINI0+0x08		; arkanoid ball difference Y
TSTX	equ	RAMINI0+0x09		; arkanoid position X for brick tests
TSTY	equ	RAMINI0+0x0A		; arkanoid position Y for brick tests
PAD		equ	RAMINI0+0x0B		; Paddle position

NEW		equ	RAMINI0+0x0C		; Rotary encoder status
OLD		equ	RAMINI0+0x0D		; Rotary encoder status

SCRINI	equ	RAMINI0+0x30		; First screen buffer location
SCR0  	equ	RAMINI0+0x30		; Screen buffer - row 1
SCR1  	equ	RAMINI0+0x31		; Screen buffer - row 1
SCR2  	equ	RAMINI0+0x32		; Screen buffer - row 1
SCR3  	equ	RAMINI0+0x33		; Screen buffer - row 1
SCR4  	equ	RAMINI0+0x34		; Screen buffer - row 1
SCR5  	equ	RAMINI0+0x35		; Screen buffer - row 1
SCR6  	equ	RAMINI0+0x36		; Screen buffer - row 1
SCREND	equ	RAMINI0+0x36		; Last screen buffer location



; --- Flags

#define		INT_OCC		FLAGS,0		; Interrupt occured
#define		BUT_PEND	FLAGS,1		; Interrupt occured
#define		ServKey		FLAGS,2		; Interrupt occured
#define		ServRE		FLAGS,3		; Interrupt occured
#define		RECW		FLAGS,4		; Interrupt occured


; ------ Program Section

		org	0x000

PowerUp:
		clrf	STATUS			; ensure we are at bank 0
		clrf	PCLATH			; ensure page bits before goto !!
		clrf	INTCON			; disable all interrupts
		goto	Start


;
;**********************************************************************************************************************
; ISR (Interrupt Service Routines)
;**********************************************************************************************************************
;
	org	0x004
Interrupt:
	movwf	INT_W	; 1 Save registers W, STATUS, PCLATH, FSR
	swapf	STATUS,w	; 2
	clrf	STATUS	; 3 Ensure port / timer access in bank 0
	movwf	INT_STAT	; 4
	movf	PCLATH,w	; 
	movwf	INT_PCLATH	; save PCLATH, interrupt uses page 0
	clrf	PCLATH	; 
	movf	FSR,w
	movwf	INT_FSR

	bcf	PIR1,TMR2IF	; open int for timer2
	clrf	CHPORT		; off all leds
	comf	NXT_TRIS,w	; prepare tri state
	bsf	STATUS,RP0	; bank 1
	movwf	CHTRIS		; upload tri state
	bcf	STATUS,RP0	; bank 0
	movf	NXT_PORT,w	; prepare port data
	movwf	CHPORT		; upload port data

	movf	BIT_CNT,w	; we are refreshing 4*per ms, then divide it
	andlw	0x03
	btfsc	STATUS,Z
	decf	MSCNT,f
	bsf		INT_OCC

	incf	BIT_CNT,f	; next bit
	movlw	0x38		; last bit is 55d=37h
	; movlw	0x10		; last bit is 16d=10h
	xorwf	BIT_CNT,w	; last bit
	btfsc	STATUS,Z
	clrf	BIT_CNT		; clear counter

	movf	BIT_CNT,w
	call	BitConvert
	movwf	NXT_PORT
	movf	BIT_CNT,w
	call	TrisConvert
	movwf	NXT_TRIS

	rlf		BIT_CNT,w	; BIT_CNT=xxAAADDD (AAA = byte DDD = bit in bite)
	andlw	0x70		; 0AAA0000
	movwf	FSR
	swapf	FSR,w		; 00000AAA
	addlw	SCRINI		; + first addr of buffer
	movwf	FSR			; into FSR
	movf	BIT_CNT,w
	call	BitConvert	; bit position of DDD
	andwf	INDF,w		; check if is 0 or 1
	btfsc	STATUS,Z	; if zero = dark
	clrf	NXT_TRIS	; disable leds
         
EndInt:
	movf	INT_FSR,w
	movwf	FSR
	movf	INT_PCLATH,w
	movwf	PCLATH
	swapf	INT_STAT,w
	movwf	STATUS
	swapf	INT_W,f
	swapf	INT_W,w
	retfie


; ----- Tables

BitConvert:				; also for conversion of portb mask
		andlw	0x07
		addwf	PCL,f		; Convert bit position to byte mask
		retlw	0x01
		retlw	0x02
		retlw	0x04
		retlw	0x08
		retlw	0x10
		retlw	0x20
		retlw	0x40
		retlw	0x80

RBitConvert:				
		andlw	0x07
		addwf	PCL,f		; Convert bit position to byte mask - in reverse
		retlw	0x80
		retlw	0x40
		retlw	0x20
		retlw	0x10
		retlw	0x08
		retlw	0x04
		retlw	0x02
		retlw	0x01

TrisConvert:
		addwf	PCL,f		; Convert position to tris mask
		retlw 0x03		;0
		retlw 0x03		;1
		retlw 0x05
		retlw 0x09
		retlw 0x11
		retlw 0x21
		retlw 0x41
		retlw 0x81
		retlw 0x05
		retlw 0x06
		retlw 0x06
		retlw 0x0A
		retlw 0x12
		retlw 0x22
		retlw 0x42
		retlw 0x82
		retlw 0x09
		retlw 0x0A
		retlw 0x0C
		retlw 0x0C
		retlw 0x14
		retlw 0x24
		retlw 0x44
		retlw 0x84
		retlw 0x11
		retlw 0x12
		retlw 0x14
		retlw 0x18
		retlw 0x18
		retlw 0x28
		retlw 0x48
		retlw 0x88
		retlw 0x21
		retlw 0x22
		retlw 0x24
		retlw 0x28
		retlw 0x30
		retlw 0x30
		retlw 0x50
		retlw 0x90
		retlw 0x41
		retlw 0x42
		retlw 0x44
		retlw 0x48
		retlw 0x50
		retlw 0x60
		retlw 0x60
		retlw 0xA0
		retlw 0x81
		retlw 0x82
		retlw 0x84
		retlw 0x88
		retlw 0x90
		retlw 0xA0
		retlw 0xC0		; 54
		retlw 0xC0		; 55
                 
        if ($ > d'255')
                ERROR "  Tables exceded page 0.   If it works, why do you change it?   "
        endif


; ----- Power Up Section ( Be sure is in first 2K program code)

Start:
		movlw	0x07			; PIC 16F628, all digital
		movwf	CMCON                          
		clrf	RCSTA			; disable UART

		movlw 	RA_INI			; Set ports
		movwf 	PORTA                         
		movlw 	RB_INI
		movwf 	PORTB
		SEL_BANK_1
		movlw	PIE1_INI
		movwf	PIE1			; peripheral interrupts
		movlw	RB_TRIS
		movwf 	TRISB
		movlw 	RA_TRIS         	; Set port A I/O configuration
		movwf 	TRISA
		movlw 	OPTION_INI      	; PORTB pull-ups disabled, Timer configuration
		movwf 	OPTION_REG
		movlw	0x0B			; internal oscillator to 4MHz
		movwf	PCON
		movlw	0x00	; VRR off voltage reference
		movwf	VRCON
		movlw	PR2_INI
		movwf	PR2

		SEL_BANK_0

		clrf	TMR0			; init timers
		movlw	T2CON_INI
		movwf	T2CON
		movlw	0x07	; RA0, RA1, RA2 digital inputs
		movwf	CMCON                          

		movlw	RAMINI0			; Clear variables bank 0 & bank 1
		movwf	FSR
ClearRAM:
		clrf	INDF
		incf	FSR,f
		btfss	FSR,7	
		goto	ClearRAM

		movlw	INTC_INI		; Set interrupts
		movwf	INTCON

		bcf	PIR1,TMR2IF	; open int for timer2
		bsf	INTCON,GIE	; enable interupt
		bsf	T2CON,TMR2ON	; enable timer2

		movlw b'00001110'
		movwf SCR0
		movlw b'00001110'
		movwf SCR1
		movlw b'00001110'
		movwf SCR2
		movlw b'00001110'
		movwf SCR3
		movlw b'00001110'
		movwf SCR4
		movlw b'00001110'
		movwf SCR5
		movlw b'00001110'
		movwf SCR6

		movlw d'250'
		movwf MSCNT

		movlw 3
		movwf BALLX
		movwf PAD
		movlw 1
		movwf BALLY
		movwf BALLDX
		movwf BALLDY

		bsf   ServRE

; --- Main Loop
MainLoop:
		movf	BALLX,w	; Put Ball to screen
		addlw	SCRINI
		movwf	FSR
		movf	BALLY,w
		call	RBitConvert
		iorwf	INDF,f
MLDelay
		btfsc	INT_OCC
		call	CheckKeys
		btfsc	ServKey
		call	ServiceKeys
		btfsc	ServRE
		call	ServiceEncoder
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	MLDelay	; wait until next timestamp

		movf	BALLX,w	; remove Ball from screen
		addlw	SCRINI
		movwf	FSR
		movf	BALLY,w
		call	RBitConvert
		xorlw	0xFF
		andwf	INDF,f

		movlw	d'200'
		movwf	MSCNT

		movf	BALLY,w		; on position 1 check paddle
		xorlw	0x01
		btfsc	STATUS,Z
		goto	CheckPaddle
		movf	BALLDX,w	; check brick in front of ball
		addwf	BALLX,w
		movwf	TSTX
		movf	BALLY,w
		movwf	TSTY
		call	CheckBrick
		btfsc	STATUS,Z
		goto	MLnoBrickOnX
		call	OffBrick
		movlw	0xFF	; Brick on X reached, reverse
		xorwf	BALLDX,f
		incf	BALLDX,f
MLnoBrickOnX:

		movf	BALLDY,w	; check brick in top of ball
		addwf	BALLY,w
		movwf	TSTY
		movf	BALLX,w
		movwf	TSTX
		call	CheckBrick
		btfsc	STATUS,Z
		goto	MLnoBrickOnY
		call	OffBrick
		movlw	0xFF	; Brick on X reached, reverse
		xorwf	BALLDY,f
		incf	BALLDY,f
MLnoBrickOnY:

		movf	BALLDY,w	; check brick in top+front of ball
		addwf	BALLY,w
		movwf	TSTY
		movf	BALLDX,w
		addwf	BALLX,w
		movwf	TSTX
		call	CheckBrick
		btfsc	STATUS,Z
		goto	MLnoBrickOnXY
		call	OffBrick
		movlw	0xFF	; Brick on X reached, reverse
		xorwf	BALLDY,f
		incf	BALLDY,f
		xorwf	BALLDX,f
		incf	BALLDX,f
		goto	MLnoBrickOnXY

CheckPaddle:
		btfss	BALLDY,7	; only when going down
		goto	MLnoBrickOnXY

		movf	PAD,w		; check paddle pixel one
		xorwf	BALLX,w
		btfsc	STATUS,Z
		goto	RevertY
		incf	PAD,w		; check paddle pixel two
		xorwf	BALLX,w
		btfsc	STATUS,Z
		goto	RevertY
		btfsc	BALLDX,7
		goto	CheckDecrease
		movf	BALLX,w
		addwf	BALLDX,w
		xorwf	PAD,w
		btfss	STATUS,Z
		goto	MLnoBrickOnXY
		goto	RevertXY
CheckDecrease:
		movf	BALLX,w
		addwf	BALLDX,w
		addwf	BALLDX,w
		xorwf	PAD,w
		btfss	STATUS,Z
		goto	MLnoBrickOnXY
		;goto	RevertXY

RevertXY:
		comf	BALLDX,f
		incf	BALLDX,f
RevertY
		comf	BALLDY,f
		incf	BALLDY,f

MLnoBrickOnXY:

		btfss	BALLDX,7	; if bit 7 is set, then decrease
		goto	MLxInc
		movf	BALLX,w
		btfss	STATUS,Z	; if non sero, then do not care
		goto	MLxOK
		;movlw	0xFF	; First X reached, reverse
		comf	BALLDX,f
		incf	BALLDX,f
MLxInc:
		movf	BALLX,w
		xorlw	0x06
		btfss	STATUS,Z	; if non sero, then do not care
		goto	MLxOK
		;movlw	0xFF	; Last X reached, reverse
		comf	BALLDX,f
		incf	BALLDX,f
MLxOK:
		btfss	BALLDY,7	; if bit 7 is set, then decrease
		goto	MLyInc
		movf	BALLY,w
		btfss	STATUS,Z	; if non sero, then do not care
		goto	MLyOK
		;movlw	0xFF	; First X reached, reverse
		comf	BALLDY,f
		incf	BALLDY,f
		goto	0

MLyInc:
		movf	BALLY,w
		xorlw	0x07
		btfss	STATUS,Z	; if non sero, then do not care
		goto	MLyOK
		;movlw	0xFF	; Last X reached, reverse
		comf	BALLDY,f
		incf	BALLDY,f
MLyOK:
		movf	BALLDX,w	; move ball
		addwf	BALLX,f
		movlw	0x06	; check for last X
		xorwf	BALLX,w
		btfss	STATUS,Z
		goto	MLxhOK
		;movlw	0xFF	; Last X reached, reverse
		comf	BALLDX,f
		incf	BALLDX,f
MLxhOK:
		movf	BALLX,w	; check for first X
		btfss	STATUS,Z
		goto	MLxlOK
		;movlw	0xFF	; First X reached, reverse
		comf	BALLDX,f
		incf	BALLDX,f
MLxlOK:
		movf	BALLDY,w
		addwf	BALLY,f
		movlw	0x07	; check for last Y
		xorwf	BALLY,w
		btfss	STATUS,Z
		goto	MLyhOK
		;movlw	0xFF	; First Y reached, reverse
		comf	BALLDY,f
		incf	BALLDY,f

MLyhOK:
		movf	BALLY,w	; check for first Y
		btfss	STATUS,Z
		goto	MLylOK
		;movlw	0xFF	; First Y reached, reverse
		;xorwf	BALLDY,f
		;incf	BALLDY,f
		movf	BALLX,w	; Put Ball to screen
		addlw	SCRINI
		movwf	FSR
		movf	BALLY,w
		call	RBitConvert
		iorwf	INDF,f

		comf	SCR0,f
		comf	SCR1,f
		comf	SCR2,f
		comf	SCR3,f
		comf	SCR4,f
		comf	SCR5,f
		comf	SCR6,f
		movlw	0x80
		movwf	MSCNT
GODelay1
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	GODelay1	; wait until next timestamp

		comf	SCR0,f
		comf	SCR1,f
		comf	SCR2,f
		comf	SCR3,f
		comf	SCR4,f
		comf	SCR5,f
		comf	SCR6,f

		movlw	0xFA
		movwf	MSCNT
GODelay2
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	GODelay2	; wait until next timestamp

		movlw	0xFA
		movwf	MSCNT
GODelay3
		movf	MSCNT,w
		btfss	STATUS,Z
		goto	GODelay3	; wait until next timestamp

		goto	0

MLylOK:
		goto MainLoop

ServiceEncoder
	bcf		ServRE
	btfss	RECW
	goto	SEncCCW
SEncCW:
	incf	PAD,f
	movlw	0x06
	xorwf	PAD,w
	btfss	STATUS,Z
	goto	SEcom
	movlw	0x05
	movwf	PAD
	goto	SEcom
SEncCCW:
	decf	PAD,f
	btfsc	PAD,7
	clrf	PAD

SEcom:
	decf	PAD,w
	movwf	TSTX
	clrf	TSTY
	btfss	TSTX,7
	call	OffBrick
	incf	TSTX,f
	call	OnBrick
	incf	TSTX,f
	call	OnBrick
	incf	TSTX,f
	incf	TSTX,w
	andlw	0x07
	btfss	STATUS,Z
	call	OffBrick
	return

ServiceKeys:
	bcf	ServKey
	return



CheckKeys:
	bcf		INT_OCC
;**********************************************************************************************************************
;Rotary Encoder Routines
;Logic is to mimic the detection of the "<", ">", and 'SEL' keys
;**********************************************************************************************************************
;   ___
;__|   |____  PORTA,2
;     ___
;____|   |__  PORTA,1
; --> CW   <-- CCW

	btfss	RE_B        ;Encoder Button has priority
	goto	Button_Pressed

	bcf		BUT_PEND        ;Clear Button Flag
	movf	PORTA,w        ;Read input port
	andlw	b'00001100'        ;leave only encoder inputs
	movwf	NEW        ;save as new value
	xorwf	OLD,w        ;compare with previous reading
	btfsc	STATUS,Z        ;will be Zero if no change
	return                ;and return
Process_Encoder:
	btfsc	OLD,2                ; --- In a case of raising edge
	goto 	EncoderEnd         ; //
	btfss	NEW,2           ; /
	goto	EncoderEnd
	bsf		RECW        ;">"
	btfss	NEW,3        ;Bit clear means clockwise rotation
	bcf		RECW        ;"<"
	bsf		ServRE        ;Save as key value
EncoderEnd:
	movf        NEW,w        ;Move NEW to OLD for next sample
	movwf        OLD
	return
Button_Pressed:
	btfsc	BUT_PEND        ;already processed?
	return                ;yes, return
	bsf		ServKey        ;indicate key needs servicing
	bsf		BUT_PEND        ;prevent multiple samples
	return                ;

CheckBrick:		; check status on position TSTX:TSTY 
		movf	TSTX,w	; Put Ball to screen
		addlw	SCRINI
		movwf	FSR
		movf	TSTY,w
		call	RBitConvert
		andwf	INDF,w
		return

OffBrick:		; destroy brick on position TSTX:TSTY 
		movf	TSTX,w	; Put Ball to screen
		addlw	SCRINI
		movwf	FSR
		movf	TSTY,w
		call	RBitConvert
		xorlw	0xFF
		andwf	INDF,f
		return

OnBrick:		; destroy brick on position TSTX:TSTY 
		movf	TSTX,w	; Put Ball to screen
		addlw	SCRINI
		movwf	FSR
		movf	TSTY,w
		call	RBitConvert
		iorwf	INDF,f
		return

Delay10us:
		DNOP				; 1,2	Delay w * 10us
		DNOP				; 3,4
		DNOP				; 5,6

		addlw	0xFF			; 7
		btfss	STATUS,Z		; 8
		goto	Delay10us		; 9,10
		return



; ---------------------- Internal EEPROM -------------------------------------

		org	0x2100

		dt	"CharliPX"
		dt	" J.Fucik"
		dt	(__VERDAY   >> 4)  +0x30
		dt	(__VERDAY   & 0x0F)+0x30,"/"
		dt	(__VERMONTH >> 4)  +0x30
		dt	(__VERMONTH & 0x0F)+0x30,"/"
		dt	(__VERYEAR  >> 4)  +0x30
		dt	(__VERYEAR  & 0x0F)+0x30

		end